import collections
import copy
import abc
import inspect
import functools
import zipfile
from typing import Union, Type, Mapping
from dryml.utils import is_nonstring_iterable, is_dictlike, \
    get_class_from_str, get_class_str, get_hashed_id, init_arg_list_handler, \
    init_arg_dict_handler, is_supported_scalar_type, is_supported_listlike, \
    is_supported_dictlike, map_dictlike, map_listlike, equal_listlike, \
    equal_dictlike
from dryml.context.context_tracker import WrongContextError, \
    context, NoContextError
from dryml.context.process import compute_context
from dryml.save_cache import SaveCache
from dryml.file_intermediary import FileIntermediary
import uuid


class MissingIdError(Exception):
    pass


class MissingMetadataError(Exception):
    pass


class IncompleteDefinitionError(Exception):
    pass


class ComputeModeAlreadyActiveError(Exception):
    pass


class ComputeModeLoadError(Exception):
    pass


class ComputeModeNotActiveError(Exception):
    pass


class ComputeModeSaveError(Exception):
    pass


class ExpectedArgumentError(Exception):
    pass


class BuildStratTracker(object):
    def __init__(self):
        self.tracker = {}

    def __getitem__(self, dry_id):
        if dry_id not in self.tracker:
            self.tracker[dry_id] = set()
        return self.tracker[dry_id]

    def __repr__(self):
        return f"{self.tracker}"


# Create a couple global variables
build_repo = None
build_cache = None
build_verbose = None
def_cache = None
build_strat = None


def is_concrete_val(input_object):
    from dryml import Object
    # Is this object a dry definition?
    if isinstance(input_object, Object):
        # A Object itself is a concrete value
        return True
    elif isinstance(input_object, ObjectDef) or \
            is_dictlike(input_object) and 'dry_def' in input_object:
        # Check that there's a Dry ID here.
        if 'dry_id' not in input_object['dry_kwargs']:
            return False
        # Check the args
        for arg in input_object['dry_args']:
            if not is_concrete_val(arg):
                return False
        # Check the kwargs
        for key in input_object['dry_kwargs']:
            kwarg = input_object['dry_kwargs'][key]
            if not is_concrete_val(kwarg):
                return False
    elif is_nonstring_iterable(input_object):
        for obj in input_object:
            if not is_concrete_val(obj):
                return False
    elif is_dictlike(input_object):
        for key in input_object:
            val = input_object[key]
            if not is_concrete_val(val):
                return False
    elif issubclass(type(input_object), type):
        # We have a type which is 'callable'
        return True
    elif callable(input_object):
        # Callables aren't concrete. They don't have a definite value
        return False
    # Assuming all other types are concrete for now.
    return True


class Meta(abc.ABCMeta):
    def __new__(cls, clsname, bases, attrs):
        # Set default init function as python's default
        # init tries to use super which shouldn't be done
        # here.
        if '__init__' not in attrs:
            attrs['__init__'] = Meta.default_init

        # Create class
        new_cls = super().__new__(cls, clsname, bases, attrs)

        # Get init function
        init_func = new_cls.__init__

        # Run check for self
        Meta.check_for_self(init_func, clsname)

        # Detect if we have the base class.
        base = False
        if '__dry_meta_base__' in attrs:
            base = True

        # Set new methods which need the class object
        # to be set properly
        new_cls.__init__ = Meta.make_dry_init(new_cls, init_func, base=base)
        new_cls.load_object = Meta.make_load_object(new_cls)
        new_cls.save_object = Meta.make_save_object(new_cls)
        new_cls.load_compute = Meta.make_load_compute(new_cls)
        new_cls.save_compute = Meta.make_save_compute(new_cls)
        new_cls.compute_prepare = Meta.make_compute_prepare(new_cls)
        new_cls.compute_cleanup = Meta.make_compute_cleanup(new_cls)

        # Wrap marked compute methods
        if hasattr(new_cls, '__dry_compute_methods__'):
            for (attr, ctx_kwargs) in new_cls.__dry_compute_methods__:
                if hasattr(new_cls, attr):
                    method = getattr(new_cls, attr)
                    if 'ctx_dont_create_context' not in ctx_kwargs:
                        ctx_kwargs['ctx_dont_create_context'] = True
                    setattr(
                        new_cls, attr,
                        compute_context(**ctx_kwargs)(method))

        return new_cls

    @staticmethod
    def check_for_self(f, clsname):
        sig = inspect.signature(f)
        keys = list(sig.parameters.keys())
        first_par = sig.parameters[keys[0]]
        if first_par.name != 'self':
            raise RuntimeError(
                "__init__ signature of Meta Objects must have a first "
                f"argument named self. Found when initializing type {clsname}")

    @staticmethod
    def track_args(f):
        """
        Build a list of explicitly mentioned arguments which will be
        consumed by dry args systems
        """
        # Get function signature
        sig = inspect.signature(f)

        # Build list of positional and keyword arguments for dry_init
        track_args = getattr(f, '__dry_args__', [])
        track_kwargs = getattr(f, '__dry_kwargs__', [])
        existing_kwargs = list(map(lambda t: t[0], track_kwargs))
        for key in sig.parameters:
            par = sig.parameters[key]
            if par.name == 'self':
                continue
            if par.kind == inspect.Parameter.VAR_POSITIONAL:
                continue
            if par.kind == inspect.Parameter.VAR_KEYWORD:
                continue
            if par.default == inspect.Parameter.empty and \
               par.kind != inspect.Parameter.KEYWORD_ONLY:
                track_args.append(par.name)
            else:
                if par.name not in existing_kwargs:
                    track_kwargs.append((par.name, par.default))

        f.__dry_args__ = track_args
        f.__dry_kwargs__ = track_kwargs

        return f

    @staticmethod
    def collect_args(f):
        f.__dry_collect_args__ = True
        return f

    @staticmethod
    def collect_kwargs(f):
        f.__dry_collect_kwargs__ = True
        return f

    @staticmethod
    def mark_compute(f):
        f.__dry_compute_mark__ = True
        return f

    @staticmethod
    def default_init(self, *args, **kwargs):
        pass

    # Create scope with __class__ defined so super can find a cell
    # with the right name.
    # Python issue: https://bugs.python.org/issue29944
    @staticmethod
    def make_dry_init(__class__, init_func, base=False):
        # Track arguments
        init_func = Meta.track_args(init_func)
        sig = inspect.signature(init_func)

        # Detect var signature components
        no_var_pars = False
        no_final_kwargs = False
        last_par_kind = None
        for key in sig.parameters:
            par = sig.parameters[key]
            last_par_kind = par.kind
            if par.kind == inspect.Parameter.VAR_POSITIONAL:
                no_var_pars = True
        if last_par_kind != inspect.Parameter.VAR_KEYWORD:
            no_final_kwargs = True

        @functools.wraps(init_func)
        def dry_init(self, *args, dry_args=None, dry_kwargs=None, **kwargs):
            # Initialize dry arguments
            dry_args = init_arg_list_handler(dry_args)
            dry_kwargs = init_arg_dict_handler(dry_kwargs)

            # Initialize compute data holder
            if not hasattr(self, '__dry_compute_data__'):
                self.__dry_compute_data__ = None

            # Initialize compute mode indicator
            if not hasattr(self, '__dry_compute_mode__'):
                self.__dry_compute_mode__ = False

            # Initialize compute context indicator
            if not hasattr(self, '__dry_compute_context__'):
                self.__dry_compute_context__ = 'default'

            # manage whether to collect all arguments
            collect_args = False
            if hasattr(init_func, '__dry_collect_args__'):
                if init_func.__dry_collect_args__:
                    collect_args = True

            # Determine how many arguments to collect
            if collect_args:
                # Collect all the arguments
                num_args = len(args)
            else:
                num_args = len(init_func.__dry_args__)

            if num_args > len(args):
                raise ExpectedArgumentError(
                    f"Object class {__class__} init expects {num_args} "
                    f"positional arguments, got {len(args)}. Did you forget "
                    f"to specify a required positional argument?")

            for i in range(num_args):
                dry_args.append(args[i])

            collect_kwargs = False
            if hasattr(init_func, '__dry_collect_kwargs__'):
                if init_func.__dry_collect_kwargs__:
                    collect_kwargs = True

            # Collect keyword arguments and save into dry_kwargs
            used_kwargs = []
            for k, v in init_func.__dry_kwargs__:
                # Need to use .get since we are passing a default (v).
                dry_kwargs[k] = kwargs.get(k, v)
                used_kwargs.append(k)

            # Collect remaining kwargs if needed.
            dont_collect_kwargs = ['dry_id', 'dry_metadata']
            if collect_kwargs:
                for k in kwargs:
                    if k not in dont_collect_kwargs:
                        if k not in used_kwargs:
                            dry_kwargs[k] = kwargs[k]
                            used_kwargs.append(k)

            # Grab unaltered arguments to pass to super
            super_args = args[num_args:]
            used_kwargs += ['dry_args', 'dry_kwargs']
            super_kwargs = {
                k: v for k, v in kwargs.items() if k not in used_kwargs
            }

            if base:
                # At the base, we need to validate the dry args
                # and dry kwargs
                validate_val_obj(dry_args)
                validate_val_obj(dry_kwargs)

                self.dry_args = tuple(dry_args)
                self.dry_kwargs = dry_kwargs

                # Construct parents
                try:
                    super().__init__(
                        *super_args,
                        **super_kwargs,
                    )
                except TypeError as e:
                    print(f"error constructing possibly 'object' parent. args:"
                          f" {super_args} kwargs: {super_kwargs}")
                    raise e
            else:
                # Construct parents
                super().__init__(
                    *super_args,
                    dry_args=dry_args,
                    dry_kwargs=dry_kwargs,
                    **super_kwargs,
                )

            # Execute user init
            # Here we make sure to remove special arguments
            used_kwargs = ['dry_args', 'dry_kwargs']
            if no_final_kwargs:
                known_kwargs = list(map(
                    lambda t: t[0],
                    init_func.__dry_kwargs__))
                sub_kwargs = {
                    k: v for k, v in kwargs.items() if k in known_kwargs}
            else:
                sub_kwargs = {
                    k: v for k, v in kwargs.items() if k not in used_kwargs}

            # Remove dry_id from being used in non-base constructors.
            # Kludge solution.
            if not base:
                if 'dry_id' in sub_kwargs:
                    del sub_kwargs['dry_id']
                if 'dry_metadata' in sub_kwargs:
                    del sub_kwargs['dry_metadata']

            if no_var_pars:
                args = args[:num_args]

            # Store list of Objects we need to later save.
            if not hasattr(self, '__dry_obj_container_list__'):
                self.__dry_obj_container_list__ = []

            def _add_dry_objs(el):
                from dryml import Object
                if isinstance(el, Object):
                    self.__dry_obj_container_list__.append(el)
                if is_nonstring_iterable(el):
                    for elm in el:
                        _add_dry_objs(elm)
                if is_dictlike(el):
                    for key in el:
                        elm = el[key]
                        _add_dry_objs(elm)

            _add_dry_objs(args)
            _add_dry_objs(sub_kwargs)

            # Call user defined init
            init_func(self, *args, **sub_kwargs)

        return dry_init

    @staticmethod
    def make_load_object(__class__):
        """
        Method for making a load_object function
        """
        def load_object(self, file: zipfile.ZipFile) -> bool:
            if not hasattr(__class__, '__dry_meta_base__'):
                # We're not the base, call the super class's load.
                if not super().load_object(file):
                    print("issue with super class load")
                    return False
            else:
                # Load compute data at the base.
                compute_data_path = 'compute_data.zip'
                if self.__dry_compute_data__ is not None:
                    del self.__dry_compute_data__
                if compute_data_path in file.namelist():
                    with file.open(compute_data_path, 'r') as f:
                        new_compute_data = FileIntermediary()
                        new_compute_data.write(f.read())
                        self.__dry_compute_data__ = new_compute_data
                else:
                    self.__dry_compute_data__ = None

            # Call this class's load object.
            if hasattr(__class__, 'load_object_imp'):
                retval = __class__.load_object_imp(self, file)
                if type(retval) is not bool:
                    raise TypeError("load_object_imp must return a bool.")
                if not retval:
                    print("issue with object load implementation")
                    return False
            return True
        return load_object

    @staticmethod
    def make_save_object(__class__):
        """
        Method for making a save_object function
        """
        def save_object(self, file: zipfile.ZipFile, save_cache=None) -> bool:
            if hasattr(__class__, '__dry_meta_base__'):
                # We're at the base, so load the compute data.
                # Save any compute data from compute components.
                # If compute mode is active
                if self.__dry_compute_mode__:
                    self.save_compute(save_cache=save_cache)

                # Save compute data if it's there
                if self.__dry_compute_data__ is not None:
                    compute_data_path = 'compute_data.zip'
                    data_buff = self.__dry_compute_data__
                    with file.open(compute_data_path, 'w') as f:
                        data_buff.write_to_file(f)

            # Call this class's save object.
            if hasattr(__class__, 'save_object_imp'):
                retval = __class__.save_object_imp(self, file)
                if type(retval) is not bool:
                    raise TypeError(
                        "save_object_imp must always return a bool.")
                if not retval:
                    return False

            if not hasattr(__class__, '__dry_meta_base__'):
                # If we're not the base, call the super class's save.
                super().save_object(file)

            # Save contained dry objects passed as arguments to construct
            # for obj in self.__dry_obj_container_list__:
            #     obj_id = obj.dry_id
            #     save_path = f'dry_objects/{obj_id}.dry'
            #     if save_path not in file.namelist():
            #         with file.open(save_path, 'w') as f:
            #             if not obj.save_self(f):
            #                 return False

            return True
        return save_object

    @staticmethod
    def make_compute_prepare(__class__):
        """
        Method for making compute_prepare function
        """

        def compute_prepare(self, _top=True):
            is_top_call = False
            if _top:
                # We know that we're the first call.
                _top = False
                is_top_call = True

                # throw error
                if self.__dry_compute_mode__:
                    raise ComputeModeAlreadyActiveError()

                # Check for context ONLY ONCE.
                # Check required context for this class
                required_context_name = \
                    getattr(self, '__dry_compute_context__', 'default')

                # Get context manager
                ctx_manager = context()

                # Check requirements against context
                if ctx_manager is None:
                    raise NoContextError("There is no context active.")
                else:
                    ctx_reqs = {required_context_name: {}}
                    if not ctx_manager.satisfies(ctx_reqs):
                        raise WrongContextError(
                            f"{ctx_manager} currently doesn't satisfy "
                            f"requirements {ctx_reqs}")

                # Prepare contained objects I would do this only at top.
                # for obj in self.__dry_obj_container_list__:
                #     obj.compute_prepare()

            # Prepare self super classes
            if not hasattr(__class__, '__dry_meta_base__'):
                # Call the super class's compute_prepare
                super().compute_prepare(_top=_top)

            # Execute user compute prepare method
            if hasattr(self, 'compute_prepare_imp'):
                self.compute_prepare_imp()

            if is_top_call:
                self.__dry_compute_mode__ = True

        return compute_prepare

    @staticmethod
    def make_compute_cleanup(__class__):
        """
        Method for making compute_cleanup function
        """

        def compute_cleanup(self):
            if not self.__dry_compute_mode__:
                raise ComputeModeNotActiveError()

            # Cleanup contained objects
            # for obj in self.__dry_obj_container_list__:
            #     obj.compute_cleanup()

            # Cleanup this object
            if hasattr(self, 'compute_cleanup_imp'):
                self.compute_cleanup_imp()

            # Cleanup parent classes
            if not hasattr(__class__, '__dry_meta_base__'):
                # If we're not the base, call the super class's save.
                super().compute_cleanup()

            self.__dry_compute_mode__ = False

        return compute_cleanup

    @staticmethod
    def make_load_compute(__class__):
        """
        Method for making a load_compute function
        """

        def load_compute(self, f=None) -> bool:
            if f is None:
                # Load contained objects
                # if f is None, we're at the top, so load here.
                # for obj in self.__dry_obj_container_list__:
                #     # We call load_compute without f since those objects
                #     # will have individual saved compute states.
                #     if not obj.load_compute():
                #         return False

                if self.__dry_compute_data__ is None:
                    # There is no data to load.
                    return True
                else:
                    # Set file pointer.
                    f = self.__dry_compute_data__

            # self.__dry_compute_data__ exists, so we need to load it.

            if not hasattr(__class__, '__dry_meta_base__'):
                # If we're not the base, call the super class's load.
                # So, we're loading from the ground up.
                super().load_compute(f=f)

            # Load this object's compute
            if hasattr(__class__, 'load_compute_imp'):
                if f is not None:
                    # Need to seek the io stream to the beginning
                    f.seek(0)
                    with zipfile.ZipFile(
                            f, mode='r') as zf:
                        imp_res = __class__.load_compute_imp(
                            self, zf)
                        if type(imp_res) is not bool:
                            raise TypeError(
                                "load_compute_imp must return a bool.")
                        if not imp_res:
                            return False
            # Finally report success
            return True
        return load_compute

    @staticmethod
    def make_save_compute(__class__):
        """
        Method for making a save_compute function
        """
        def save_compute(self, f=None, save_cache=None) -> bool:
            if save_cache is None:
                save_cache = SaveCache()
            top_call = False
            if f is None:
                top_call = True
                # We're at the top of the call stack.
                # Create io bytes stream
                f = FileIntermediary()

                # Save contained dry objects passed as arguments to construct
                for obj in self.__dry_obj_container_list__:
                    if not obj.save_compute(save_cache=save_cache):
                        return False

            # Now we check if we're in the save cache.
            if id(self) in save_cache.compute_cache:
                # We don't have to go any further.
                return True

            # Call class save implementation
            if hasattr(__class__, 'save_compute_imp'):
                with zipfile.ZipFile(
                        f, mode='w') as zf:
                    compute_imp_res = __class__.save_compute_imp(
                        self, zf)
                    if type(compute_imp_res) is not bool:
                        raise TypeError("save_compute_imp must return a bool.")
                    if not compute_imp_res:
                        return False

            # Call super class save
            if not hasattr(__class__, '__dry_meta_base__'):
                # If we're not the base, call the super class's load.
                super().save_compute(f=f, save_cache=save_cache)

            if top_call:
                # Only set the save file if there's data.
                if not f.is_empty():
                    self.__dry_compute_data__ = f
                else:
                    # delete the io stream created.
                    del f

                # Add self to the save cache, so we don't repeat.
                save_cache.compute_cache.add(id(self))

            return True
        return save_compute


def validate_key(key):
    if type(key) in (str, bytes, int, float):
        return
    if type(key) is tuple:
        return
    raise TypeError(f"Unsupported key ({key}) of type {type(key)}")


# Assumption, for Objects, definitions are not valid values.
# All definitions should be resolved into objects.
def validate_val_obj(val):
    if is_supported_scalar_type(val):
        return
    from dryml import Object
    if isinstance(val, Object):
        # Assumption: Objects have already validated their values
        return
    if isinstance(val, ObjectDef):
        raise TypeError(
            "Object Definitions not valid for use within Object")
    if is_supported_listlike(val):
        for el in val:
            validate_val_obj(el)
        return
    if is_supported_dictlike(val):
        for k in val:
            validate_val_obj(val[k])
        return
    raise ValueError(
        f"value ({val}) of type {type(val)} not supported")


def validate_val_def(val):
    if is_supported_scalar_type(val):
        return
    from dryml import Object
    if isinstance(val, Object):
        # Assumption: Objects have already validated their values
        return
    if isinstance(val, ObjectDef):
        return
    if is_supported_listlike(val):
        for el in val:
            validate_val_def(el)
        return
    if is_supported_dictlike(val):
        for k in val:
            validate_val_def(val[k])
        return
    raise ValueError(
        f"value ({val}) of type {type(val)} not supported")


def strip_dry_id(obj, render_cache=None):
    # Create render cache
    if render_cache is None:
        render_cache = {}

    # Check render cache if this obj has been seen before
    obj_id = id(obj)
    if obj_id in render_cache:
        return render_cache[obj_id]

    # Construct new object
    if is_dictlike(obj):
        new_dict = {}
        for k in obj:
            if k != 'dry_id':
                new_dict[k] = strip_dry_id(obj[k], render_cache=render_cache)
        render_cache[obj_id] = new_dict
        return new_dict
    elif is_nonstring_iterable(obj):
        new_list = [strip_dry_id(o) for o in obj]
        if type(obj) is tuple:
            return tuple(new_list)
        else:
            return new_list
    else:
        return obj


class RenderCache(object):
    def __init__(self):
        self.unique_cache = {}
        self.nonunique_cache = {}


def def_to_obj(val, repo=None, load_zip=None):
    def applier(val):
        return def_to_obj(val, repo=repo, load_zip=load_zip)
    from dryml import Object
    if is_supported_scalar_type(val):
        return val
    elif isinstance(val, ObjectDef):
        return val.build(repo=repo, load_zip=load_zip)
    elif isinstance(val, Object):
        return val
    elif is_supported_listlike(val):
        return map_listlike(applier, val)
    elif is_supported_dictlike(val):
        return map_dictlike(applier, val)
    else:
        raise TypeError(
            f"Unsupported value {val} of type {type(val)} encountered!")


def def_to_cat_def(val, cache=None):
    def applier(val):
        return def_to_cat_def(val, cache=cache)
    from dryml import Object
    if is_supported_scalar_type(val):
        return val
    elif isinstance(val, ObjectDef):
        return val.get_cat_def(recursive=True, cache=cache)
    elif isinstance(val, Object):
        return val.definition().get_cat_def(recursive=True, cache=cache)
    elif is_supported_listlike(val):
        return map_listlike(applier, val)
    elif is_supported_dictlike(val):
        return map_dictlike(applier, val)
    else:
        raise TypeError(
            f"Unsupported value {val} of type {type(val)} encountered!")


class ObjectDef(collections.UserDict):
    @staticmethod
    def from_dict(def_dict: Mapping, render_cache=None):
        raise RuntimeError("Functionality Questionable")
        # Construct cache object if needed
        if render_cache is None:
            render_cache = RenderCache()

        # Check render cache
        if 'dry_id' in def_dict['dry_kwargs']:
            # Check for unique
            dry_id = def_dict['dry_kwargs']['dry_id']
            if dry_id in render_cache.unique_cache:
                return render_cache.unique_cache[dry_id]
        else:
            # Check for nonunique
            dict_id = id(def_dict)
            if dict_id in render_cache.nonunique_cache:
                return render_cache.nonunique_cache[dict_id]

        # We have to construct a new definition.
        def transform_el(el):
            if type(el) is dict:
                if 'dry_def' in el:
                    # Detect dry def.
                    return ObjectDef.from_dict(
                        el, render_cache=render_cache)
            if is_dictlike(el):
                # We have just a normal dictionary.
                return {k: transform_el(el[k]) for k in el}
            elif is_nonstring_iterable(el):
                new_list = [transform_el(v) for v in el]
                if type(el) is tuple:
                    return tuple(new_list)
                else:
                    return new_list
            else:
                return el

        cls = def_dict['cls']
        args = transform_el(def_dict.get('dry_args', ()))
        kwargs = transform_el(def_dict.get('dry_kwargs', {}))

        # construct new definition
        new_def = ObjectDef(
            cls,
            *args,
            dry_mut=def_dict.get('dry_mut', False),
            **kwargs)

        # Save our result in the cache.
        if 'dry_id' in def_dict['dry_kwargs']:
            dry_id = def_dict['dry_kwargs']['dry_id']
            render_cache.unique_cache[dry_id] = new_def
        else:
            def_id = id(def_dict)
            render_cache.nonunique_cache[def_id] = new_def

        # Return the result.
        return new_def

    def __init__(self, cls: Union[Type, str],
                 *args, dry_mut: bool = False, **kwargs):
        self._tracking_id = uuid.uuid4()

        super().__init__()
        if cls is None:
            raise ValueError(
                "Can't construct an object Definition with None type.")
        self.cls = cls
        self.data['dry_mut'] = dry_mut
        # Validate arguments
        validate_val_def(args)
        self.data['dry_args'] = args
        validate_val_def(kwargs)
        self.data['dry_kwargs'] = kwargs

    def __eq__(self, other):
        return self.equal(other)

    @staticmethod
    def equal_func(val_a, val_b, cls_str_compare=False):
        if isinstance(val_a, ObjectDef) and \
                not isinstance(val_b, ObjectDef):
            return False
        if not isinstance(val_a, ObjectDef) and \
                isinstance(val_b, ObjectDef):
            return False
        if isinstance(val_a, ObjectDef):
            return val_a.equal(val_b, cls_str_compare=cls_str_compare)
        else:
            return (val_a == val_b)

    def equal(self, other, cls_str_compare=True):
        def eq_func(val_a, val_b):
            return ObjectDef.equal_func(
                val_a, val_b, cls_str_compare=cls_str_compare)

        if isinstance(other, type(self)):
            if cls_str_compare:
                self_cls = get_class_str(self['cls'])
                other_cls = get_class_str(other['cls'])
                if self_cls != other_cls:
                    return False
            else:
                if self['cls'] != other['cls']:
                    return False
            if not equal_listlike(eq_func, self.args, other.args):
                return False
            if not equal_dictlike(eq_func, self.kwargs, other.kwargs):
                return False
            return True
        elif is_dictlike(other):
            # We don't support comparison directly with dictionaries for now.
            return False
        else:
            # Not same type. Give False.
            return False

    @property
    def cls(self):
        return self['cls']

    @cls.setter
    def cls(self, val: Union[Type, str]):
        self['cls'] = val

    @property
    def dry_mut(self):
        return self['dry_mut']

    @property
    def args(self):
        return self['dry_args']

    @property
    def kwargs(self):
        return self['dry_kwargs']

    @property
    def dry_id(self):
        if 'dry_id' not in self['dry_kwargs']:
            raise MissingIdError()
        return self['dry_kwargs']['dry_id']

    @property
    def tracking_id(self):
        return self._tracking_id

    def __setitem__(self, key, value):
        if key not in ['cls', 'dry_mut', 'dry_args', 'dry_kwargs']:
            raise ValueError(
                f"Setting Key {key} not supported by ObjectDef")

        if key == 'cls':
            if type(value) is type or issubclass(type(value), abc.ABCMeta):
                self.data['cls'] = value
            elif type(value) is str:
                self.data['cls'] = get_class_from_str(value)
            else:
                raise TypeError(
                    f"Value of type {type(value)} not supported "
                    "for class assignment!")
        else:
            self.data[key] = value

    def to_dict(self, cls_str: bool = False, render_cache=None):
        raise RuntimeError("Functionality Questionable")
        from dryml import Object

        # Create cache if it doesn't yet exist.
        if render_cache is None:
            render_cache = RenderCache()

        # Search for a rendered version of this definition first.
        if self.is_concrete():
            dry_id = self.dry_id
            if dry_id in render_cache.unique_cache:
                return render_cache.unique_cache[dry_id]
        else:
            def_id = id(self)
            if def_id in render_cache.nonunique_cache:
                return render_cache.nonunique_cache[def_id]

        # Create recursive helper function
        def transform_el(el):
            if type(el) is ObjectDef:
                return el.to_dict(
                    cls_str=cls_str, render_cache=render_cache)
            elif type(el) is Object:
                the_def = el.definition()
                return the_def.to_dict(
                    cls_str=cls_str, render_cache=render_cache)
            elif is_dictlike(el):
                # May cause a problem if we have a dict version
                # of a ObjectDef within another ObjectDef
                return {k: transform_el(el[k]) for k in el}
            elif is_nonstring_iterable(el):
                new_list = [transform_el(e) for e in el]
                if type(el) is tuple:
                    return tuple(new_list)
                else:
                    return new_list
            else:
                return el

        # Create the new dictionary
        new_dict = {
            'cls': self.cls if not cls_str else get_class_str(self.cls),
            'dry_mut': self.dry_mut,
            'dry_args': transform_el(self.args),
            'dry_kwargs': transform_el(self.kwargs),
            'dry_def': True,
        }

        # Save result to cache
        if self.is_concrete():
            render_cache.unique_cache[self.dry_id] = new_dict
        else:
            def_id = id(self)
            render_cache.nonunique_cache[def_id] = new_dict

        # Return result
        return new_dict

    def build(self, repo=None, load_zip=None, verbose=None):
        "Construct an object"
        reset_repo = False
        reset_cache = False
        reset_strat = False
        reset_verbose = False
        construct_object = True

        global build_repo
        global build_cache
        global def_cache
        global build_strat
        global build_verbose

        # Handle setting of build_verbose
        if build_verbose is None:
            if verbose is None:
                build_verbose = False
            else:
                if type(verbose) is not bool:
                    raise TypeError("verbose must be a bool!")
                build_verbose = verbose
            reset_verbose = True
        else:
            if verbose is not None:
                if type(verbose) is not bool:
                    raise TypeError("verbose must be a bool!")
                if verbose != build_verbose:
                    raise ValueError(
                        "Can't change verbose once set by a superior call.")

        # Handle creation of build repo
        if repo is not None:
            if build_repo is not None:
                if repo is not build_repo:
                    raise RuntimeError(
                        "different repos not currently supported")
            else:
                # Set the call_repo
                build_repo = repo
                reset_repo = True

        # Handle creation of build cache
        if build_cache is None:
            build_cache = {}
            def_cache = {}
            reset_cache = True

        # Handle creation of build strat cache
        if build_strat is None:
            build_strat = BuildStratTracker()
            reset_strat = True

        # Define a cleanup function to call in the event of error
        # and at the end of the function.
        def cleanup():
            global build_repo
            global build_cache
            global def_cache
            global build_strat
            global build_verbose

            # Reset the repo for this function
            if reset_repo:
                build_repo = None

            # Reset the build cache
            if reset_cache:
                build_cache = None
                def_cache = None

            # Reset the build strat cache
            if reset_strat:
                build_strat = None

            # Reset the verbose indicator
            if reset_verbose:
                build_verbose = None

        # Create some book-keeping variables
        obj = None
        construction_required = True
        construct_object = True

        # Check whether this SPECIFIC definition has been built yet.
        if self.tracking_id in def_cache:
            obj = def_cache[self.tracking_id]
            if build_verbose:
                print(
                    "Object with id {obj.dry_id} built for definition "
                    f"with tracking id {self.tracking_id} was found "
                    "in the definition cache.")
            construction_required = False
            construct_object = False

        # Indicates whether we need to construct the object because this
        # definition isn't concrete
        if obj is None and self.is_concrete():
            # This is a concrete object and we can get an id.
            obj_id = self.dry_id
            construction_required = False

        # Check the cache
        if obj is None and not construction_required and \
                build_cache is not None:
            try:
                obj = build_cache[obj_id]
                if build_verbose:
                    print(f"Found object with id {obj_id} in "
                          "the build cache.")
                construct_object = False
            except KeyError:
                pass

        # Check the repo
        if obj is None and (not construction_required) and \
                (build_repo is not None) and \
                ('repo' not in build_strat[obj_id]):
            try:
                build_strat[obj_id].add('repo')
                obj = build_repo.get_obj(self, load=True)
                build_cache[obj_id] = obj
                def_cache[self.tracking_id] = obj
                construct_object = False
                build_strat[obj_id].remove('repo')
                if build_verbose:
                    print(f"Found object with id {obj_id} in the "
                          "repository.")
            except KeyError:
                # Didn't find the object in the repo
                pass

        try:
            # Check the zipfile
            if obj is None and (not construction_required) \
                    and construct_object and (load_zip is not None) \
                    and ('zip' not in build_strat[obj_id]):
                target_filename = f"dry_objects/{obj_id}.dry"
                from dryml import load_object
                if target_filename in load_zip.namelist():
                    build_strat[obj_id].add('zip')
                    with load_zip.open(target_filename) as f:
                        obj = load_object(f)
                        build_cache[obj_id] = obj
                        def_cache[self.tracking_id] = obj
                        construct_object = False
                    build_strat[obj_id].remove('zip')
                    if build_verbose:
                        print(f"Found object with id {obj_id} in the "
                              "zip file.")

            # Finally, actually construct the object
            if obj is None and construct_object:
                new_args = def_to_obj(self.args, repo=repo, load_zip=load_zip)
                new_kwargs = def_to_obj(
                    self.kwargs,
                    repo=repo,
                    load_zip=load_zip)

                obj = self.cls(*new_args, **new_kwargs)

                # Save object in the build cache.
                obj_id = obj.dry_id
                build_cache[obj_id] = obj
                def_cache[self.tracking_id] = obj
                if build_verbose:
                    print(f"Explicitly constructed new object. id: {obj_id}")

            elif obj is None and not construct_object:
                raise RuntimeError(
                    "Unexpected condition encountered when "
                    "building from definition")

        except Exception as e:
            cleanup()
            raise e

        cleanup()

        # Return the result
        return obj

    def get_cat_def(self, recursive=True, cache=None):
        if recursive:
            # Create cache if needed
            if cache is None:
                cache = {}

            # check the cache for whether this definition has been rendered yet
            if id(self) in cache:
                return cache[id(self)]

            # Recursively apply get_cat_def to args, kwargs
            new_args = def_to_cat_def(self.args, cache=cache)
            new_kwargs = def_to_cat_def(self.kwargs, cache=cache)
            if 'dry_id' in new_kwargs:
                del new_kwargs['dry_id']
            if 'dry_metadata' in new_kwargs:
                del new_kwargs['dry_metadata']

            # Create new definition
            cat_def = ObjectDef(
                self.cls, *new_args, dry_mut=self.dry_mut, **new_kwargs)

            # Store the result
            cache[id(self)] = cat_def

            # return the result
            return cat_def

        else:
            # Remove only the top level dry_id.
            new_def = ObjectDef(
                self.cls, *self.args, dry_mut=self.dry_mut, **self.kwargs)
            del new_def.kwargs['dry_id']
            return new_def

    def get_hash_str(self, no_id: bool = False, no_metadata: bool = False):
        class_hash_str = get_class_str(self.cls)
        args_hash_str = str(self.args)
        # Remove dry_id so we can test for object 'class'
        if no_id or no_metadata:
            kwargs_copy = copy.copy(self.kwargs)
            if no_id:
                if 'dry_id' in kwargs_copy:
                    kwargs_copy.pop('dry_id')
            if no_metadata:
                if 'dry_metadata' in kwargs_copy:
                    kwargs_copy.pop('dry_metadata')
            kwargs_hash_str = str(kwargs_copy)
        else:
            kwargs_hash_str = str(self.kwargs)
        return class_hash_str+args_hash_str+kwargs_hash_str

    def is_concrete(self):
        return is_concrete_val(self)

    def __hash__(self):
        if not self.is_concrete():
            raise IncompleteDefinitionError(
                "Definition {self} has no dry_id!")
        return hash(self.get_individual_id())

    def get_individual_id(self):
        if not self.is_concrete():
            raise IncompleteDefinitionError(
                "Definition {self} has no dry_id!")
        return get_hashed_id(self.get_hash_str())

    def get_category_id(self):
        return get_hashed_id(self.get_hash_str(no_id=True, no_metadata=True))



from dryml.file_intermediary import FileIntermediary
import zipfile
import pytest
import pickle
import os
from dryml.utils import pickler


@pytest.mark.usefixtures("create_name")
def test_file_intermediary_1(create_name):
    # Create intermediary
    int_file = FileIntermediary()

    test_file = 'test.txt'
    test_text = "TEST"

    # Create test zip as usual
    z_file = zipfile.ZipFile(int_file, mode='w')
    with z_file.open("test.txt", 'w') as f:
        f.write(test_text.encode('utf-8'))
    z_file.close()

    # Write to the temp location
    with open(create_name, 'wb') as f:
        int_file.write_to_file(f)
    int_file.close()

    # Read temp file into zipfile.
    with open(create_name, 'rb') as f:
        z_file = zipfile.ZipFile(f, mode='r')
        assert test_file in z_file.namelist()
        with z_file.open(test_file, 'r') as f2:
            assert f2.read().decode('utf-8') == test_text


@pytest.mark.usefixtures("create_name")
def test_file_intermediary_2(create_name):
    # Create intermediary
    int_file = FileIntermediary()

    test_file = 'test.txt'
    test_text = "TEST"

    # Create test zip as usual
    z_file = zipfile.ZipFile(int_file, mode='w')
    with z_file.open("test.txt", 'w') as f:
        f.write(test_text.encode('utf-8'))
    z_file.close()

    # Write to the temp location and close intermediary
    int_file.write_to_file(create_name)
    int_file.close()

    # Read temp file into zipfile.
    with open(create_name, 'rb') as f:
        z_file = zipfile.ZipFile(f, mode='r')
        assert test_file in z_file.namelist()
        with z_file.open(test_file, 'r') as f2:
            f2.read().decode('utf-8') == test_text


@pytest.mark.usefixtures("create_name")
def test_file_intermediary_3(create_name):
    # Create intermediary
    int_file = FileIntermediary(mem_mode=True)

    test_file = 'test.txt'
    test_text = "TEST"

    # Create test zip as usual
    z_file = zipfile.ZipFile(int_file, mode='w')
    with z_file.open("test.txt", 'w') as f:
        f.write(test_text.encode('utf-8'))
    z_file.close()

    # Write to the temp location
    with open(create_name, 'wb') as f:
        int_file.write_to_file(f)
    int_file.close()

    # Read temp file into zipfile.
    with open(create_name, 'rb') as f:
        z_file = zipfile.ZipFile(f, mode='r')
        assert test_file in z_file.namelist()
        with z_file.open(test_file, 'r') as f2:
            f2.read().decode('utf-8') == test_text


@pytest.mark.usefixtures("create_name")
def test_file_intermediary_4(create_name):
    # Create intermediary
    int_file = FileIntermediary(mem_mode=True)

    test_file = 'test.txt'
    test_text = "TEST"

    # Create test zip as usual
    z_file = zipfile.ZipFile(int_file, mode='w')
    with z_file.open("test.txt", 'w') as f:
        f.write(test_text.encode('utf-8'))
    z_file.close()

    # Write to the temp location and close intermediary
    int_file.write_to_file(create_name)
    int_file.close()

    # Read temp file into zipfile.
    with open(create_name, 'rb') as f:
        z_file = zipfile.ZipFile(f, mode='r')
        assert test_file in z_file.namelist()
        with z_file.open(test_file, 'r') as f2:
            f2.read().decode('utf-8') == test_text


@pytest.mark.usefixtures("create_name")
def test_file_intermediary_5(create_name):
    int_file = FileIntermediary()
    z_file = zipfile.ZipFile(int_file, mode='w')

    # Save meta data analogue
    meta_data = {
        'version': 1
    }

    meta_dump = pickler(meta_data)
    with z_file.open('meta_data.pkl', mode='w') as f:
        f.write(meta_dump)

    z_file.close()
    int_file.write_to_file(create_name)
    int_file.close()

    # Load meta data analogue
    with open(create_name, 'rb') as file:
        z_file = zipfile.ZipFile(file, mode='r')

        with z_file.open('meta_data.pkl', 'r') as meta_file:
            assert pickle.loads(meta_file.read()) == meta_data


@pytest.mark.usefixtures("create_name")
def test_file_intermediary_6(create_name):
    # Create intermediary
    int_file = FileIntermediary()

    test_file = 'test.txt'
    test_text = "TEST"

    # Create test zip as usual
    z_file = zipfile.ZipFile(int_file, mode='w')
    with z_file.open("test.txt", 'w') as f:
        f.write(test_text.encode('utf-8'))
    z_file.close()

    # Write to the temp location
    with open(create_name, 'wb') as f:
        int_file.write_to_file(f)

    # Read temp file into zipfile.
    with open(create_name, 'rb') as f:
        z_file = zipfile.ZipFile(f, mode='r')
        assert test_file in z_file.namelist()
        with z_file.open(test_file, 'r') as f2:
            assert f2.read().decode('utf-8') == test_text

    # Remove file
    os.remove(create_name)

    # Write to temp location again.
    with open(create_name, 'wb') as f:
        int_file.write_to_file(f)

    # Read temp file into zipfile.
    with open(create_name, 'rb') as f:
        z_file = zipfile.ZipFile(f, mode='r')
        assert test_file in z_file.namelist()
        with z_file.open(test_file, 'r') as f2:
            assert f2.read().decode('utf-8') == test_text

    int_file.close()



from dryml.data import Dataset
from dryml.models import Trainable
from dryml.context import compute_context
import numpy as np


@compute_context(ctx_dont_create_context=True)
def mean_squared_error(trainable: Trainable, test_data: Dataset):
    if not test_data.supervised:
        raise ValueError("Dataset is unsupervised!")

    # Prepare input data, expects a dataset
    data = trainable.eval(test_data) \
                    .as_not_indexed()

    if not data.batched:
        data = data.batch()

    data = data.numpy()

    total_loss = 0.
    num_examples = 0
    for batch_e_y, batch_y in data:
        total_loss += np.sum((batch_e_y-batch_y)**2)
        num_examples += batch_e_y.shape[0]

    return total_loss/num_examples


@compute_context(ctx_dont_create_context=True)
def categorical_accuracy(trainable: Trainable, test_data: Dataset):
    if not test_data.supervised:
        raise ValueError("Dataset is unsupervised!")

    # Prepare the input data, expects a dataset
    data = trainable.eval(test_data) \
                    .as_not_indexed()

    if not data.batched:
        data = data.batch()

    num_correct = 0
    num_total = 0

    for Y_eval, Y in data.numpy():
        num_correct += (Y_eval == Y).sum()
        num_total += len(Y)

    return num_correct/num_total



import zipfile
import pickle
from dryml.object import Object
from dryml.context import cls_method_compute


@cls_method_compute('train')
@cls_method_compute('eval')
@cls_method_compute('prep_train', ctx_use_existing_context=True)
@cls_method_compute('prep_eval', ctx_use_existing_context=True)
class Trainable(Object):
    untrained = 0
    trained = 2

    def __init__(self, *args, **kwargs):
        self.train_state = Trainable.untrained

    def load_object_imp(self, file: zipfile.ZipFile) -> bool:
        # Load parent components first
        with file.open('component_data.pkl', 'r') as f:
            component_data = pickle.load(f)
        self.train_state = component_data['train_state']
        return True

    def save_object_imp(self, file: zipfile.ZipFile) -> bool:
        with file.open('component_data.pkl', 'w') as f:
            f.write(pickle.dumps({'train_state': self.train_state}))
        return True

    def prep_train(self):
        # Configure the model for training
        pass

    def prep_eval(self):
        # Configure the model for evaluation
        pass

    def train(self, *args, train_spec=None, train_callbacks=[]):
        # Handle the setting of the train state flag
        self.train_state = Trainable.trained
        # This should be the last step in training so no more super is needed

    def eval(self, data, *args, **kwargs):
        raise NotImplementedError("Method not defined for a base Trainable")



from __future__ import annotations
from typing import Callable
from dryml.data.util import nestize


class NotIndexedError():
    pass


class NotSupervisedError():
    pass


class Dataset(object):
    """
    A Simple wrapper class to house data operations
    """

    def __init__(self, indexed=False, supervised=False,
                 batch_size=None):
        self._indexed = indexed
        self._supervised = supervised
        self._batch_size = batch_size

    @property
    def indexed(self) -> bool:
        """
        Indicate whether this dataset is indexed.
        """
        return self._indexed

    def index(self):
        """
        If indexed, return the index of this dataset
        """
        if not self.indexed:
            raise NotIndexedError()

        return self.map(lambda t: t[0])

    def as_indexed(self, start=0) -> Dataset:
        """
        If not already indexed, return a version of this dataset
        which is indexed.
        """
        raise NotImplementedError()

    def as_not_indexed(self):
        """
        Strip index from dataset
        """
        if not self.indexed:
            return self
        else:
            return self.map(lambda t: t[1])

    @property
    def supervised(self) -> bool:
        """
        Indicate whether this dataset is supervised (has targets as well)
        """
        return self._supervised

    def as_not_supervised(self) -> Dataset:
        """
        Strip supervised targets
        """

        if not self.supervised:
            return self
        else:
            if self.indexed:
                return self.map(lambda t: (t[0], t[1][0]))
            else:
                return self.map(lambda t: t[0])

    def intersect(self) -> Dataset:
        """
        Intersect this dataset with another
        """
        raise NotImplementedError()

    @property
    def data_gen(self):
        """
        Gives a function where calling it returns a generator of the dataset.
        """
        raise NotImplementedError()

    def data(self):
        """
        Return the backing dataset
        """
        raise NotImplementedError()

    @property
    def batched(self) -> bool:
        """
        Indicate whether this data has been batched
        """
        if self._batch_size is not None:
            return True
        else:
            return False

    @property
    def batch_size(self):
        """
        Get the batch size
        """

        return self._batch_size

    def batch(self, batch_size=32) -> Dataset:
        """
        Batch this data
        """
        raise NotImplementedError()

    def unbatch(self) -> Dataset:
        """
        Unbatch this data
        """
        raise NotImplementedError()

    def map(self, func: Callable = None) -> Dataset:
        """
        Apply a function to the data of Dataset
        """
        raise NotImplementedError()

    def map_el(self, func: Callable = None) -> Dataset:
        """
        Apply a function to every element in Dataset, even nesting in
        """

        return self.map(nestize(func))

    def apply_X(self, func: Callable = None) -> Dataset:
        """
        Apply a function to the X component of Dataset
        """

        if self.indexed:
            if self.supervised:
                return self.map(lambda t: (t[0], (func(t[1][0]), t[1][1])))
            else:
                return self.map(lambda t: (t[0], func(t[1])))
        else:
            if self.supervised:
                return self.map(lambda t: (func(t[0]), t[1]))
            else:
                return self.map(lambda x: func(x))

    def apply_Y(self, func=None) -> Dataset:
        """
        Apply a function to the Y component of Dataset
        """

        if not self.supervised:
            raise NotSupervisedError(
                "Can't apply a function to the Y component of "
                "non supervised dataset")

        if self.indexed:
            return self.map(lambda i, xy: (i, (xy[0], func(xy[1]))))
        else:
            return self.map(lambda x, y: (x, func(y)))

    def apply(self, func=None) -> Dataset:
        """
        Apply a function to (X, Y)
        """

        if not self.supervised:
            raise NotSupervisedError(
                "Can't apply a function to the Y component of "
                "non supervised dataset")

        if self.indexed:
            return self.map(lambda i, xy: (i, func(*xy)))
        else:
            return self.map(lambda x, y: func(x, y))

    def __iter__(self):
        """
        Create iterator
        """

        return iter(self.data())

    def take(self, n):
        """
        Take only a specific number of examples
        """
        raise NotImplementedError()

    def skip(self, n):
        """
        Skip a specific number of examples
        """
        raise NotImplementedError()

    def __len__(self):
        """
        Get length of dataset. Will return Infinite if infinite,
        and unknown if it can't be determined.
        """
        raise NotImplementedError()

    def numpy(self):
        """
        Create a NumpyDataset from this dataset.
        """

        raise NotImplementedError()

    def tf(self):
        """
        Create a TFDataset from this dataset.
        """

        raise NotImplementedError()

    def collect(self):
        """
        Collect all data from the dataset
        """

        result = []
        for el in self:
            result.append(el)
        return result

    def peek(self):
        """
        Get the first element to have a look
        """

        item_list = self.take(1).collect()

        if len(item_list) == 0:
            raise RuntimeError(
                "Can't peek, no data in dataset. If you expect data, "
                "double check there isn't a batch function called "
                "with drop_remainder=True.")

        return item_list[0]

    def count(self, limit=-1):
        """
        Attempt to count 'elements' in the Dataset
        """

        number = 0
        for e in self:
            number += 1
            if limit > 0:
                if number > limit:
                    break
        return number

    def shuffle(self, buffer_size, seed=None):
        """
        Shuffle elements of dataset.
        """

        raise NotImplementedError()



import pathlib


class dryml_object_saver(object):
    def __init__(self, obj, checkpoint_dir_base, train_state):
        self.obj = obj
        self.checkpoint_dir_base = checkpoint_dir_base
        self.train_state = train_state

    def __call__(self):
        checkpoint_dir = \
            f"{self.checkpoint_dir_base}_{self.train_state.global_step()}"
        pathlib.Path(checkpoint_dir).mkdir(parents=True, exist_ok=True)
        self.obj.save_self(f"{checkpoint_dir}/model.dry")
        self.train_state.save(f"{checkpoint_dir}/train_state.pkl")



from dryml.config import ObjectDef, Meta, \
    IncompleteDefinitionError, ComputeModeAlreadyActiveError, \
    ComputeModeLoadError, ComputeModeNotActiveError, \
    ComputeModeSaveError, MissingIdError
from dryml.object import Object, ObjectFile, ObjectFactory, \
    load_object, save_object, change_object_cls, \
    ObjectWrapper, CallableWrapper, get_contained_objects, \
    build_obj_tree
from dryml.selector import Selector
from dryml.repo import Repo
from dryml.collections import List, Tuple, Dict
from dryml.workshop import Workshop
from dryml.context import compute_context, compute
import dryml.context as context

__version__ = "0.2.0"

__all__ = [
    Object,
    ObjectFile,
    ObjectDef,
    Meta,
    ObjectFactory,
    Selector,
    Repo,
    List,
    Tuple,
    Dict,
    ObjectWrapper,
    CallableWrapper,
    Workshop,
    load_object,
    save_object,
    change_object_cls,
    context,
    IncompleteDefinitionError,
    ComputeModeAlreadyActiveError,
    ComputeModeLoadError,
    ComputeModeNotActiveError,
    ComputeModeSaveError,
    MissingIdError,
    build_obj_tree,
    compute_context,
    compute,
    get_contained_objects,
]



from typing import Optional, Union
from dryml.context import ComputeContext
from dryml.context.context_tracker import ResourceRequest


class TorchComputeContext(ComputeContext):
    def __init__(
            self,
            resource_request: Optional[Union[ResourceRequest, dict]] = {}):
        super().__init__(resource_request=resource_request)

    def acquire_context(self):
        # Let parent handle allocation
        super().acquire_context()

        # Get allocated gpus
        alloc_gpus = self.allocation.gpus

        # check if we need to set memory limits on any of these gpus
        need_mem_limit = False
        for gpu_key in alloc_gpus:
            if self.allocation[gpu_key] < 1.:
                need_mem_limit = True
                break

        if need_mem_limit:
            print("WARNING, currently there is no way to "
                  "limit memory used by torch.")

    def compute_devices(self):
        device_list = list(map(
            lambda n: n.replace('gpu/', 'cuda:'),
            self.allocation.gpus))
        if len(device_list) == 0:
            device_list = ['cpu']
        return device_list

    def release_context(self):
        # Let parent handle releasing allocation
        super().release_context()



from dryml.data import Dataset, \
    NumpyDataset, util
from typing import Callable
import tensorflow as tf
import numpy as np


class TFDataset(Dataset):
    def __init__(
            self, in_ds: tf.data.Dataset, indexed=False,
            supervised=False, batch_size=None, size=None):
        super().__init__(
            indexed=indexed, supervised=supervised,
            batch_size=batch_size)
        self.ds = in_ds
        if size is None:
            size = np.nan
        self.size = size

    def as_indexed(self, start=0) -> Dataset:
        """
        If not already indexed, return a version of this dataset
        which is indexed.
        """
        if self.indexed:
            return self
        else:
            return TFDataset(
                self.ds.enumerate(start=start),
                indexed=True,
                supervised=self.supervised,
                batch_size=self.batch_size)

    @property
    def data_gen(self):
        return lambda: self.ds

    def data(self):
        """
        Get the internal dataset
        """
        return self.ds

    def batch(self, batch_size=32, drop_remainder=True) -> Dataset:
        """
        Batch this data
        """
        if self.batched:
            if self.batch_size == batch_size:
                return self
            else:
                return self.unbatch().batch(batch_size=batch_size)
        else:
            return TFDataset(
                self.ds.batch(
                    batch_size=batch_size,
                    drop_remainder=drop_remainder),
                indexed=self.indexed,
                supervised=self.supervised,
                batch_size=batch_size)

    def unbatch(self) -> Dataset:
        """
        Unbatch this data
        """
        if not self.batched:
            return self
        else:
            return TFDataset(
                self.ds.unbatch(),
                indexed=self.indexed,
                supervised=self.supervised)

    def map(self, func: Callable = None) -> Dataset:
        """
        Apply a function to the X component of Dataset
        """

        return TFDataset(
            self.ds.map(
                lambda *t: func(t),
                num_parallel_calls=tf.data.AUTOTUNE),
            indexed=self.indexed,
            supervised=self.supervised,
            batch_size=self.batch_size,
            size=self.size)

    def take(self, n):
        """
        Take only a specific number of examples
        """
        return TFDataset(
            self.ds.take(n),
            indexed=self.indexed,
            supervised=self.supervised,
            batch_size=self.batch_size)

    def skip(self, n):
        """
        Skip a specific number of examples
        """
        return TFDataset(
            self.ds.skip(n),
            indexed=self.indexed,
            supervised=self.supervised,
            batch_size=self.batch_size)

    def __len__(self):
        """
        Get length of dataset. Will return Infinite if infinite,
        and unknown if it can't be determined.
        """
        cardinality = self.ds.cardinality()
        if cardinality == tf.data.INFINITE_CARDINALITY:
            return np.inf
        if cardinality == tf.data.UNKNOWN_CARDINALITY:
            return np.nan
        return cardinality.numpy()

    def numpy(self):
        """
        Produce NumpyDataset from this TFDataset
        """

        def numpy_transform(el):
            if type(el) is not np.ndarray:
                return el.numpy()
            else:
                return el

        def numpy_generator():
            return map(
                util.nestize(numpy_transform),
                self.data_gen())

        return NumpyDataset(
            numpy_generator,
            indexed=self.indexed,
            supervised=self.supervised,
            batch_size=self.batch_size)

    def tf(self):
        """
        Get a TFDataset, but this is already a TFDataset
        """
        return self

    def torch(self):
        """
        Create TorchDataset from this dataset
        """

        import torch
        from dryml.data.torch import TorchDataset, TorchIterableDatasetWrapper

        def tf_to_torch(el):
            return torch.tensor(el.numpy())

        obj = TorchIterableDatasetWrapper(
            lambda: map(util.nestize(tf_to_torch), self.data_gen()))

        return TorchDataset(
            obj,
            indexed=self.indexed,
            supervised=self.supervised,
            batch_size=self.batch_size,
            size=self.size)

    def shuffle(self, buffer_size, seed=None):
        return TFDataset(
            self.ds.shuffle(buffer_size, seed=seed),
            indexed=self.indexed,
            supervised=self.supervised,
            batch_size=self.batch_size)



from dryml.models.sklearn.base import Model, \
    ClassifierModel, RegressionModel, BasicTraining, \
    Trainable

__all__ = [
    Model,
    ClassifierModel,
    RegressionModel,
    BasicTraining,
    Trainable,
]



from dryml.models import Trainable
from dryml.data.dataset import Dataset
import tensorflow as tf
import inspect


def func_source_extract(func):
    # Get the source code
    code_string = inspect.getsource(func)

    # Possibly strip leading spaces
    code_lines = code_string.split('\n')

    init_strip = code_lines[0]

    for line in code_lines[1:]:
        i = 0
        while i < len(init_strip) and \
                i < len(line) and \
                init_strip[i] == line[i]:
            i += 1
        if i == len(init_strip) or \
                i == len(line):
            continue

        init_strip = init_strip[:i]

    if len(init_strip) > 0:
        code_lines = list(map(
            lambda l: l[len(init_strip):],
            code_lines))

    return '\n'.join(code_lines)


class FuncXMap(Trainable):
    @staticmethod
    def from_function(func, *args, **kwargs):
        return FuncXMap(func_source_extract(func), *args, **kwargs)

    def __init__(self, func_code, *args, **kwargs):
        # Evaluate passed function code
        lcls = {}
        exec(func_code, globals(), lcls)

        # Check for function definition
        if len(lcls) == 0:
            raise ValueError("Code defines no objects!")
        if len(lcls) > 1:
            raise ValueError("Code defines more than one object!")

        # Get newly defined object
        func = list(lcls.values())[0]

        if not callable(func):
            raise ValueError(
                "Function code doesn't contain a function definition!")

        self.func = func
        self.train_state = Trainable.trained

    def train(self, *args, train_spec=None, **kwargs):
        # We can't train a plain function
        pass

    def eval(self, data: Dataset, *args, **kwargs):
        return data.tf().apply_X(func=self.func)


class FuncYMap(Trainable):
    @staticmethod
    def from_function(func, *args, **kwargs):
        return FuncYMap(func_source_extract(func), *args, **kwargs)

    def __init__(self, func_code, *args, **kwargs):
        # Evaluate passed function code
        lcls = {}
        exec(func_code, globals(), lcls)

        # Check for function definition
        if len(lcls) == 0:
            raise ValueError("Code defines no objects!")
        if len(lcls) > 1:
            raise ValueError("Code defines more than one object!")

        # Get newly defined object
        func = list(lcls.values())[0]

        if not callable(func):
            raise ValueError(
                "Function code doesn't contain a function definition!")

        self.func = func
        self.train_state = Trainable.trained

    def train(self, *args, train_spec=None, **kwargs):
        # We can't train a plain function
        pass

    def eval(self, data: Dataset, *args, **kwargs):
        return data.tf().apply_Y(func=self.func)


class FuncMap(Trainable):
    @staticmethod
    def from_function(func, *args, **kwargs):
        return FuncMap(inspect.getsource(func), *args, **kwargs)

    def __init__(self, func_code, *args, **kwargs):
        # Evaluate passed function code
        lcls = {}
        exec(func_code, globals(), lcls)

        # Check for function definition
        if len(lcls) == 0:
            raise ValueError("Code defines no objects!")
        if len(lcls) > 1:
            raise ValueError("Code defines more than one object!")

        # Get newly defined object
        func = list(lcls.values())[0]

        if not callable(func):
            raise ValueError(
                "Function code doesn't contain a function definition!")

        self.func = func
        self.train_state = Trainable.trained

    def train(self, *args, train_spec=None, **kwargs):
        # We can't train a plain function
        pass

    def eval(self, data: Dataset, *args, **kwargs):
        return data.tf().apply(func=self.func)


class BestCat(Trainable):
    def __init__(self):
        self.train_state = Trainable.trained

    def train(self, *args, train_spec=None, **kwargs):
        pass

    def eval(self, data: Dataset, *args, **kwargs):
        if not data.batched:
            data = data.batch()
        return data.tf().apply_X(
            func=lambda x: tf.argmax(x, axis=-1))


def flattener(x):
    return tf.reshape(x, [tf.shape(x)[0], -1])


class Flatten(Trainable):
    def __init__(self):
        self.train_state = Trainable.trained

    def train(self, *args, train_spec=None, **kwargs):
        pass

    def eval(self, data: Dataset, *args, **kwargs):
        if not data.batched:
            data = data.batch()
        return data.tf().apply_X(
            func=flattener)



import pickle
import zipfile
import numpy as np
from dryml.config import Meta
from dryml.models import Component, Trainable
from dryml.models import TrainFunction as BaseTrainFunction
from dryml.data import Dataset


class Model(Component):
    @Meta.collect_kwargs
    def __init__(self, cls, **kwargs):
        # It is subclass's responsibility to fill this
        # attribute with an actual keras class
        self.mdl = None
        self.cls = cls
        self.mdl_kwargs = kwargs

    def compute_prepare_imp(self):
        self.mdl = self.cls(**self.mdl_kwargs)

    def compute_cleanup_imp(self):
        self.mdl = None

    def load_compute_imp(self, file: zipfile.ZipFile) -> bool:
        pkl_file_name = 'model.pkl'
        # Load Weights
        if pkl_file_name not in file.namelist():
            # No model pickle file right now
            return True
        else:
            with file.open(pkl_file_name, 'r') as f:
                self.mdl = pickle.loads(f.read())
        return True

    def save_compute_imp(self, file: zipfile.ZipFile) -> bool:
        # Save Weights
        if self.mdl is not None:
            pkl_file_name = 'model.pkl'
            with file.open(pkl_file_name, 'w') as f:
                f.write(pickle.dumps(self.mdl))

        return True

    def __call__(self, X, *args, target=True, index=False, **kwargs):
        raise NotImplementedError()


class ClassifierModel(Model):
    def __call__(self, X, *args, target=True, index=False, **kwargs):
        return self.mdl.predict_proba(X, *args, **kwargs)


class RegressionModel(Model):
    def __call__(self, X, *args, target=True, index=False, **kwargs):
        return self.mdl.predict(X, *args, **kwargs)


class BasicTraining(BaseTrainFunction):
    """
    The basic sklearn training method.
    """

    def __init__(
            self, num_examples=-1, shuffle=False,
            shuffle_seed=None, shuffle_buffer_size=None):
        self.num_examples = num_examples
        self.shuffle = shuffle
        self.shuffle_seed = shuffle_seed
        self.shuffle_buffer_size = shuffle_buffer_size
        self.train_args = ()
        self.train_kwargs = {}

    def __call__(
            self, trainable, train_data,
            train_spec=None, train_callbacks=[]):

        train_data = train_data.unbatch()

        total_examples = self.num_examples

        if not train_data.supervised:
            raise ValueError(
                "Dataset must be supervised for basic sklearn training")

        data_examples = total_examples
        if total_examples == -1:
            data_examples = train_data.count()
            if data_examples is np.nan or data_examples is np.inf:
                raise ValueError(
                    "Dataset has no known length and you must provide "
                    "an explicit number of examples for training.")

        if self.shuffle:
            train_data = train_data.shuffle(
                self.shuffle_buffer_size,
                seed=self.shuffle_seed)

        train_data = train_data.batch(batch_size=data_examples) \
                               .as_not_indexed() \
                               .numpy()

        x, y = train_data.peek()

        trainable.mdl.fit(x, y, *self.train_args, **self.train_kwargs)


class Trainable(Trainable):
    __dry_compute_context__ = 'default'

    def __init__(self, model=None, train_fn=None, **kwargs):
        # It is subclass's responsibility to fill this
        # attribute with an actual keras class
        if model is None:
            raise ValueError("Must give a model object")
        self.model = model
        if train_fn is None:
            raise ValueError("Must give a train_fn object")
        self.train_fn = train_fn

    def eval(self, data: Dataset, *args, eval_batch_size=32, **kwargs):
        if data.batched:
            # We can execute the method directly on the data
            return data.numpy().apply_X(
                func=lambda X: self.model(X, *args, **kwargs))
        else:
            # We first need to batch the data, then unbatch to leave
            # The dataset character unchanged.
            return data.numpy().batch(batch_size=eval_batch_size) \
                       .apply_X(
                            func=lambda X: self.model(X, *args, **kwargs)) \
                       .unbatch()

    def train(
            self, data, train_spec=None, train_callbacks=[]):

        self.train_fn(self.model, data, train_spec=train_spec)



from dryml.models.tf.base import TrainFunction, Model, \
    Trainable, Wrapper
import dryml.models.tf.keras as keras
import dryml.models.tf.utils as utils

__all__ = [
    TrainFunction,
    Model,
    Trainable,
    Wrapper,
    utils,
    keras,
]



from dryml.data.dataset import Dataset
from dryml.data.util import nested_batcher, nested_unbatcher, \
    nested_flatten
from dryml.utils import is_iterator
from dryml.data.util import taker, skiper
import numpy as np
from typing import Callable


class NumpyDataset(Dataset):
    """
    A Numpy based dataset based on a list of numpy elements
    """

    def __init__(
            self, data, indexed=False,
            supervised=False, batch_size=None, size=None):

        if type(data) is np.ndarray or type(data) is tuple:
            data_size = len(data)
            if type(data) is tuple:
                size_set = set(map(lambda d: len(d), nested_flatten(data)))
                if len(size_set) > 1:
                    raise ValueError(
                        "nested elements have different numbers of elements!")
                data_size = size_set.pop()
            super().__init__(
                indexed=indexed, supervised=supervised,
                batch_size=data_size)

            self._data_gen = lambda: [data]
            self.size = 1

        elif callable(data):
            # We have a method which is supposed to yield
            # A generator.
            super().__init__(
                indexed=indexed, supervised=supervised,
                batch_size=batch_size)
            self._data_gen = data
            if size is None:
                size = np.nan
            self.size = size

        elif is_iterator(data):
            # Can't use consumable iterator
            raise TypeError(
                "Can't use a consumable like an iterator or generator. "
                "Pass a callable which produces it instead.")
        else:
            df_test = False
            try:
                import pandas as pd
                if type(data) is pd.core.frame.DataFrame:
                    df_test = True
            except ImportError:
                pass
            if df_test:
                if indexed is False:
                    super().__init__(
                        indexed=indexed, supervised=supervised,
                        batch_size=len(data))
                    self._data_gen = lambda: [data.to_numpy()]
                    self.size = 1
                elif indexed is True:
                    super().__init__(
                        indexed=indexed, supervised=supervised,
                        batch_size=len(data))
                    self._data_gen = lambda: [(data.index.to_numpy(),
                                              data.to_numpy())]
                    self.size = 1
            elif type(data) is list:
                data_size = len(data)
                super().__init__(
                    indexed=indexed, supervised=supervised,
                    batch_size=batch_size)

                self._data_gen = lambda: data
                if size is None:
                    self.size = data_size
                else:
                    if size != data_size:
                        ValueError("Detected incorrect dataset size")
                    self.size = size
            else:
                super().__init__(
                    indexed=indexed, supervised=supervised,
                    batch_size=batch_size)

                self._data_gen = lambda: data
                if size is None:
                    self.size = np.nan
                else:
                    self.size = size

    def as_indexed(self, start=0) -> Dataset:
        """
        If not already indexed, return a version of this dataset
        which is indexed.
        """
        if self.indexed:
            return self
        else:
            if not self.batched:
                def enumerate_dataset(gen_func, start=0):
                    i = start
                    it = iter(gen_func())
                    while True:
                        try:
                            d = next(it)
                        except StopIteration:
                            return
                        yield (i, d)
                        i += 1

                return NumpyDataset(
                    lambda: enumerate_dataset(self.data_gen, start=start),
                    indexed=True,
                    supervised=self.supervised,
                    batch_size=self.batch_size,
                    size=self.size)
            else:
                def enumerate_dataset(gen_func, start=0):
                    it = iter(gen_func())
                    i = start
                    while True:
                        try:
                            d = next(it)
                        except StopIteration:
                            return
                        from dryml.data.util import get_data_batch_size
                        batch_size = get_data_batch_size(d)
                        idx = np.array(list(range(i, i+batch_size)))
                        i += batch_size
                        yield (idx, d)

                return NumpyDataset(
                    lambda: enumerate_dataset(self.data_gen, start=start),
                    indexed=True,
                    supervised=self.supervised,
                    batch_size=self.batch_size,
                    size=self.size)

    @property
    def data_gen(self):
        """
        Get a function which returns a generator which yields the dataset
        """
        return self._data_gen

    def data(self):
        """
        Get the internal dataset
        """
        return self.data_gen()

    def batch(self, batch_size=32, drop_remainder=True) -> Dataset:
        """
        Batch this data
        """
        if self.batched:
            if self.batch_size != batch_size:
                return self.unbatch().batch(batch_size=batch_size)
            else:
                return self
        else:
            return NumpyDataset(
                lambda: nested_batcher(
                     self.data_gen,
                     batch_size,
                     lambda e: np.stack(e, axis=0),
                     drop_remainder=drop_remainder),
                indexed=self.indexed,
                supervised=self.supervised,
                batch_size=batch_size,
                size=self.size)

    def unbatch(self) -> Dataset:
        """
        Unbatch this data
        """
        if not self.batched:
            return self
        else:
            return NumpyDataset(
                lambda: nested_unbatcher(self.data_gen),
                indexed=self.indexed,
                supervised=self.supervised,
                size=self.size)

    def map(self, func: Callable = None) -> Dataset:
        """
        Map a function across all elements of a dataset
        """

        return NumpyDataset(
            lambda: map(func, self),
            indexed=self.indexed,
            supervised=self.supervised,
            batch_size=self.batch_size,
            size=self.size)

    def take(self, n):
        """
        Take only a specific number of examples
        """

        new_size = self.size
        if new_size is np.nan:
            new_size = n
        elif new_size is np.inf:
            new_size = n
        else:
            if new_size > n:
                new_size = n

        return NumpyDataset(
            lambda: taker(self.data_gen, n),
            indexed=self.indexed,
            supervised=self.supervised,
            batch_size=self.batch_size,
            size=new_size,
            )

    def skip(self, n):
        """
        Skip a specific number of examples
        """

        new_size = self.size
        if new_size is not np.nan and new_size is not np.inf:
            if n > new_size:
                new_size = 0
            else:
                new_size -= n

        return NumpyDataset(
            lambda: skiper(self.data_gen, n),
            indexed=self.indexed,
            supervised=self.supervised,
            batch_size=self.batch_size,
            size=new_size)

    def __len__(self):
        """
        Get length of dataset. Will return Infinite if infinite,
        and unknown if it can't be determined.
        """
        return self.size

    def numpy(self):
        return self

    def tf(self):
        from dryml.data.tf import TFDataset
        import dryml.data.util as util
        import tensorflow as tf

        # Heuristic to determine output_signature
        peek_data = self.peek()

        def get_numpy_array_spec(e):
            e_tens = tf.constant(e)
            e_spec = tf.TensorSpec.from_tensor(e_tens)
            if self.batched:
                # We need to remove the first shape number
                # Since this data is batched
                list_shape = list(e_spec.shape)
                list_shape[0] = None
                e_spec = tf.TensorSpec(
                    tf.TensorShape(list_shape),
                    e_spec.dtype)
            return e_spec
        peek_data_signature = util.nested_apply(
            peek_data, get_numpy_array_spec)

        # Create tf dataset
        dataset = tf.data.Dataset.from_generator(
            lambda: self.data_gen(),
            output_signature=peek_data_signature)

        return TFDataset(
            dataset,
            indexed=self.indexed,
            supervised=self.supervised,
            batch_size=self.batch_size,
            size=self.size)

    def torch(self):
        from dryml.data.torch import TorchIterableDatasetWrapper, TorchDataset
        import torch

        dataset = self.map_el(lambda el: torch.tensor(el))

        # Create torch dataset
        ds = TorchIterableDatasetWrapper(
            dataset.data_gen)

        return TorchDataset(
            ds,
            indexed=self.indexed,
            supervised=self.supervised,
            batch_size=self.batch_size,
            size=self.size)

    def shuffle(self, buffer_size, seed=None):
        # create generator of data
        def shuffler():
            # Create new generator
            rng = np.random.default_rng(seed=seed)

            # create iterator on unbatched data
            ds_iter = iter(self.unbatch())

            # create buffer and fill it
            el_buffer = []

            # Fill the buffer
            for idx in range(buffer_size):
                try:
                    el_buffer.append(next(ds_iter))
                except StopIteration:
                    break

            while True:
                # Compute size of buffer, and break out if nothing left
                num_in_buffer = len(el_buffer)
                if num_in_buffer == 0:
                    break

                if num_in_buffer > 1:
                    idx = rng.integers(low=0, high=num_in_buffer-1)
                else:
                    idx = 0

                yield el_buffer.pop(idx)

                # Try to refill the buffer here
                try:
                    el_buffer.append(next(ds_iter))
                except StopIteration:
                    pass

        return NumpyDataset(
            shuffler,
            indexed=self.indexed,
            supervised=self.supervised,
            batch_size=None)



from dryml.dry_repo import DryRepo
import matplotlib.pyplot as plt


def box_and_whisker(repo: DryRepo, func, func_args=None, func_kwargs=None,
                    selector_dict={}, fig_kwargs=None, **kwargs):
    if fig_kwargs is None:
        fig_kwargs = {}

    results = []
    labels = []

    for label in selector_dict:
        selector = selector_dict[label]
        scalar_results = repo.apply(
            func, func_args, func_kwargs,
            selector=selector,
            **kwargs)
        if scalar_results is None or len(scalar_results) == 0:
            print(f"WARNING: No models for label {label}, skipping.")
        else:
            results.append(scalar_results)
            labels.append(label)

    if len(results) == 0:
        print("ERROR! No models found for any label!")
        return

    # Create Figure
    fig_kwargs['figsize'] = fig_kwargs.get('figsize', (10, 10))
    fig, axes = plt.subplots(1, 1, **fig_kwargs)

    # Creating box plot
    axes.boxplot(results, labels=labels)

    # Show plots
    plt.show(fig)

    # Close Figure so we don't pollute the sessions with lots of open Figures
    plt.close(fig)



from dryml.repo import Repo
from dryml.models.trainable import Trainable


class Workshop(object):
    def __init__(self, *args, work_dir=None, create_work_dir=True, **kwargs):
        super().__init__(**kwargs)
        self.repo = Repo(directory=work_dir, create=create_work_dir)

    def data_prep(self):
        raise RuntimeError("Not implemented for base workshop")

    def train_trainable(self, obj: Trainable, *args, **kwargs):
        obj.train(self.train_data, *args, **kwargs)

    def train_trainable_outer(self, obj: Trainable, *args, **kwargs):
        if obj.train_state == Trainable.untrained:
            self.train_trainable(obj, *args, **kwargs)

    def train_models(self, *args,
                     selector=None,
                     sel_args=None, sel_kwargs=None,
                     load_objects: bool = True,
                     only_loaded: bool = False,
                     **kwargs):
        self.repo.apply(self.train_trainable_outer,
                        func_args=args,
                        func_kwargs=kwargs,
                        selector=selector,
                        sel_args=sel_args,
                        sel_kwargs=sel_kwargs,
                        load_objects=load_objects,
                        only_loaded=only_loaded)



from dryml.models.tf.keras.base import Model, Trainable, \
    BasicTraining, BasicEarlyStoppingTraining, ModelWrapper, \
    keras_sequential_functional_class, SequentialFunctionalModel

__all__ = [
    Model,
    ModelWrapper,
    Trainable,
    BasicTraining,
    BasicEarlyStoppingTraining,
    keras_sequential_functional_class,
    SequentialFunctionalModel,
]



from dryml.models import Trainable
from dryml.data.dataset import Dataset
from dryml.data.util import nestize
from dryml.data.numpy_dataset import NumpyDataset
import numpy as np
from typing import Callable


class StaticTransform(Trainable):
    def __init__(self, mode='X'):
        self.train_state = Trainable.trained
        if mode not in ['all', 'X', 'Y']:
            raise ValueError(f"mode '{mode}' not supported.")
        self.mode = mode

    def train(self, *args, train_spec=None, **kwargs):
        pass

    def applier(self, data: Dataset, func: Callable):
        if self.mode == 'all':
            return data.apply(func)
        elif self.mode == 'X':
            return data.apply_X(func)
        elif self.mode == 'Y':
            return data.apply_Y(func)
        else:
            raise RuntimeError("Unknown mode!")

    def numpy_eval(self, data: NumpyDataset):
        raise NotImplementedError()

    def eval(self, data: Dataset, *args, **kwargs):
        # Special case for Tensorflow datasets
        try:
            import dryml.data.tf
            if type(data) is dryml.data.tf.TFDataset:
                if hasattr(self, 'tf_eval'):
                    return self.tf_eval(data, *args, **kwargs)
        except ImportError:
            pass

        # Special case for Torch datasets
        try:
            import dryml.data.torch
            if type(data) is dryml.data.torch.TorchDataset:
                if hasattr(self, 'torch_eval'):
                    return self.torch_eval(data, *args, **kwargs)
        except ImportError:
            pass

        # Fallback, try to cast first to numpy then apply.
        if type(data) is not NumpyDataset:
            data = data.numpy()
        if type(data) is NumpyDataset:
            return self.numpy_eval(data, *args, **kwargs)


class BestCat(StaticTransform):
    def numpy_eval(self, data, *args, **kwargs):
        return self.applier(
            data,
            lambda x: np.argmax(x, axis=-1))

    def tf_eval(self, data, *args, **kwargs):
        import tensorflow as tf
        return self.applier(
            data,
            lambda x: tf.argmax(x, axis=-1))

    def torch_eval(self, data, *args, **kwargs):
        import torch
        return self.applier(
            data,
            lambda x: torch.argmax(x, dim=-1))


class Flatten(StaticTransform):
    def numpy_eval(self, data, *args, **kwargs):
        if data.batched:
            return self.applier(
                data,
                lambda x: x.reshape([x.shape[0], -1]))
        else:
            return self.applier(
                data,
                lambda x: x.flatten())

    def tf_eval(self, data, *args, **kwargs):
        import tensorflow as tf
        if data.batched:
            return self.applier(
                data,
                lambda x: tf.reshape(x, [tf.shape(x)[0], -1]))
        else:
            return self.applier(
                data,
                lambda x: tf.reshape(x, [-1]))

    def torch_eval(self, data, *args, **kwargs):
        import torch
        if data.batched:
            return self.applier(
                data,
                lambda x: torch.reshape(x, (torch.shape[0], -1)))
        else:
            return self.applier(
                data,
                lambda x: torch.reshape(x, (-1,)))


class Transpose(StaticTransform):
    def __init__(self, axes=None):
        self.train_state = Trainable.trained
        self.axes = axes

    def numpy_eval(self, data, *args, **kwargs):
        if data.batched:
            # Move axes up by one
            new_axes = []
            for i in self.axes:
                new_axes.append(i+1)

            return self.applier(
                data,
                lambda x: np.transpose(x, [0]+new_axes))
        else:
            return self.applier(
                data,
                lambda x: np.transpose(x, self.axes))

    def tf_eval(self, data, *args, **kwargs):
        import tensorflow as tf
        if data.batched:
            new_axes = []
            for i in self.axes:
                new_axes.append(i+1)
            new_axes = tuple(new_axes)

            return self.applier(
                data,
                lambda x: tf.transpose(x, (0,)+new_axes))
        else:
            return self.applier(
                data,
                lambda x: tf.transpose(x, self.axes))

    def torch_eval(self, data, *args, **kwargs):
        if data.batched:
            new_axes = []
            for i in self.axes:
                new_axes.append(i+1)

            return self.applier(
                data,
                lambda x: x.permute(*new_axes))
        else:
            return self.applier(
                data,
                lambda x: x.permute(*self.axes))


class Cast(StaticTransform):
    def __init__(self, dtype='float32'):
        self.dtype = dtype

    def numpy_eval(self, data, *args, **kwargs):
        np_dtype = getattr(np, self.dtype)

        def caster(x):
            return x.astype(np_dtype)

        return self.applier(
            data,
            nestize(caster))

    def tf_eval(self, data, *args, **kwargs):
        import tensorflow as tf
        tf_dtype = getattr(tf, self.dtype)

        def caster(x):
            return tf.cast(x, tf_dtype)

        return self.applier(
            data,
            nestize(caster))

    def torch_eval(self, data, *args, **kwargs):
        import torch
        torch_dtype = getattr(torch, self.dtype)

        def caster(x):
            return x.to(torch_dtype)

        return self.applier(
            data,
            nestize(caster))



import dryml
import os
import objects


def test_selector_1():
    "Class selection"
    obj1 = objects.TestClassA(base_msg="Test1", item=5)
    obj2 = objects.TestClassB([1, 2, 3], base_msg="Test1")

    sel = dryml.Selector(cls=objects.TestClassA)

    # Test selectors work with built classes
    assert sel(obj1)
    assert sel(obj1.definition())
    assert not sel(obj2)
    assert not sel(obj2.definition())

    dryml.save_object(obj1, 'test1')
    dryml.save_object(obj2, 'test2')

    # Test selectors work with loaded classes

    obj1_loaded = dryml.load_object('test1')
    obj2_loaded = dryml.load_object('test2')

    assert sel(obj1_loaded)
    assert not sel(obj2_loaded)

    os.remove('test1.dry')
    os.remove('test2.dry')


def test_selector_2():
    "args selection"
    obj1 = objects.TestClassB(1, base_msg="Test1")
    obj2 = objects.TestClassB([1, 2, 3], base_msg="Test2")

    sel = dryml.Selector(cls=objects.TestClassB, args=(1,))

    assert sel(obj1)
    assert sel(obj1.definition())
    assert not sel(obj2)
    assert not sel(obj2.definition())


def test_selector_3():
    "kwargs selection"
    obj1 = objects.TestClassA(base_msg="Test1", item='a')
    obj2 = objects.TestClassA(base_msg="Test2", item=[10, 10, 10])

    sel = dryml.Selector(
        cls=objects.TestClassA,
        kwargs={'item': 'a'})

    assert sel(obj1)
    assert sel(obj1.definition())
    assert not sel(obj2)
    assert not sel(obj2.definition())

    sel = dryml.Selector(
        cls=objects.TestClassA,
        kwargs={'item': [10, 10, 10]})

    assert not sel(obj1)
    assert not sel(obj1.definition())
    assert sel(obj2)
    assert sel(obj2.definition())


def test_selector_4():
    "superclass selection"
    obj1 = objects.TestClassA(base_msg="Test1", item='a')
    obj2 = objects.TestClassA(base_msg="Test2",
                              item=[10, 10, 10])
    obj3 = objects.TestClassB(0, base_msg="Test3")
    obj4 = objects.TestClassB([10, 10], base_msg="Test4")
    obj5 = objects.HelloInt(msg=5)
    obj6 = objects.HelloInt(msg=20)
    obj7 = objects.HelloStr(msg='test')
    obj8 = objects.HelloStr(msg='2test')

    sel = dryml.Selector(
        cls=objects.TestBase)

    assert sel(
        obj1,
        cls_str_compare=False,
        verbosity=2)
    assert sel(
        obj1.definition(),
        cls_str_compare=False,
        verbosity=2)
    assert sel(
        obj2,
        cls_str_compare=False,
        verbosity=2)
    assert sel(
        obj2.definition(),
        cls_str_compare=False,
        verbosity=2)
    assert sel(
        obj3,
        cls_str_compare=False,
        verbosity=2)
    assert sel(
        obj3.definition(),
        cls_str_compare=False,
        verbosity=2)
    assert sel(
        obj4,
        cls_str_compare=False,
        verbosity=2)
    assert sel(
        obj4.definition(),
        cls_str_compare=False,
        verbosity=2)
    assert not sel(
        obj5,
        cls_str_compare=False,
        verbosity=2)
    assert not sel(
        obj5.definition(),
        cls_str_compare=False,
        verbosity=2)
    assert not sel(
        obj6,
        cls_str_compare=False,
        verbosity=2)
    assert not sel(
        obj6.definition(),
        cls_str_compare=False,
        verbosity=2)
    assert not sel(
        obj7,
        cls_str_compare=False,
        verbosity=2)
    assert not sel(
        obj7.definition(),
        cls_str_compare=False,
        verbosity=2)
    assert not sel(
        obj8,
        cls_str_compare=False,
        verbosity=2)
    assert not sel(
        obj8.definition(),
        cls_str_compare=False,
        verbosity=2)

    sel = dryml.Selector(
        cls=objects.HelloObject)

    assert not sel(
        obj1,
        cls_str_compare=False)
    assert not sel(
        obj1.definition(),
        cls_str_compare=False)
    assert not sel(
        obj2,
        cls_str_compare=False)
    assert not sel(
        obj2.definition(),
        cls_str_compare=False)
    assert not sel(
        obj3,
        cls_str_compare=False)
    assert not sel(
        obj3.definition(),
        cls_str_compare=False)
    assert not sel(
        obj4,
        cls_str_compare=False)
    assert not sel(
        obj4.definition(),
        cls_str_compare=False)
    assert sel(
        obj5,
        cls_str_compare=False)
    assert sel(
        obj5.definition(),
        cls_str_compare=False)
    assert sel(
        obj6,
        cls_str_compare=False)
    assert sel(
        obj6.definition(),
        cls_str_compare=False)
    assert sel(
        obj7,
        cls_str_compare=False)
    assert sel(
        obj7.definition(),
        cls_str_compare=False)
    assert sel(
        obj8,
        cls_str_compare=False)
    assert sel(
        obj8.definition(),
        cls_str_compare=False)


def test_selector_5():
    """
    Nested class selection
    """

    obj = objects.TestNest(objects.HelloTrainableD(A=objects.TestNest(10)))
    obj_def = obj.definition()

    sel = dryml.Selector(
        cls=objects.TestNest,
        args=obj_def['dry_args'])

    assert sel(obj)

    del obj_def['dry_args'][0]['dry_kwargs']['dry_id']

    sel = dryml.Selector(
        cls=objects.TestNest,
        args=obj_def['dry_args'])

    assert sel(obj)

    obj_def = obj.definition()
    del obj_def['dry_args'][0]['dry_kwargs']['A']['dry_kwargs']['dry_id']

    assert sel(obj)


def test_selector_6():
    """
    A Selector usage pattern
    """

    obj = objects.TestNest(objects.HelloTrainableD(A=objects.TestNest(10)))

    obj_def = dryml.ObjectDef(
        objects.TestNest,
        dryml.ObjectDef(
            objects.HelloTrainableD,
            A=dryml.ObjectDef(
                objects.TestNest,
                10)
        )
    )

    sel = dryml.Selector.build(obj_def)

    assert sel(obj)


def test_selector_7():
    """
    Test real life example
    """

    import dryml.data.tf

    # Create transform object
    def test_func(x):
        return x*2

    real_obj = dryml.data.tf.transforms.FuncXMap.from_function(test_func)

    obj_def = dryml.ObjectDef(
        dryml.data.tf.transforms.FuncXMap,
        real_obj.dry_args[0])

    assert dryml.Selector.from_def(obj_def)(real_obj)


def test_selector_8():
    """
    Test parent/child class selection
    """

    parent_obj = objects.TestBase()

    assert dryml.Selector(objects.TestBase)(parent_obj)
    assert not dryml.Selector(objects.TestClassA)(parent_obj)

    assert dryml.Selector.from_def(
        dryml.ObjectDef(
            objects.TestBase))(parent_obj)
    assert not dryml.Selector.from_def(
        dryml.ObjectDef(
            objects.TestClassA))(parent_obj)


def test_selector_9():
    """
    Test parent/child class selection in an argument
    """

    parent_obj = objects.TestNest(objects.TestBase)

    assert dryml.Selector.from_def(
       dryml.ObjectDef(
           objects.TestNest,
           objects.TestBase))(parent_obj)
    assert not dryml.Selector.from_def(
       dryml.ObjectDef(
           objects.TestNest,
           objects.TestClassA))(parent_obj)


def test_selector_10():
    """
    Test parent/child class selection in a keyword argument
    """

    parent_obj = objects.TestNest2(A=objects.TestBase)

    assert dryml.Selector.from_def(
       dryml.ObjectDef(
           objects.TestNest2,
           A=objects.TestBase))(parent_obj)
    assert not dryml.Selector.from_def(
       dryml.ObjectDef(
           objects.TestNest2,
           A=objects.TestClassA))(parent_obj)


def test_selector_build_1():
    """
    Test that we can construct Selectors from various objects
    """

    obj = objects.TestNest(objects.HelloTrainableD(A=objects.TestNest(10)))

    sel = dryml.Selector.from_obj(obj)

    assert sel(obj)


def test_selector_build_2():
    """
    Test that we can construct Selectors from various objects
    """

    obj = objects.TestNest(objects.HelloTrainableD(A=objects.TestNest(10)))

    sel = dryml.Selector.from_def(obj.definition())

    assert sel(obj)


def test_selector_build_5():
    """
    Test that we can construct Selectors from nested objects
    """

    obj1 = objects.TestNest2(A=1)

    obj_def = dryml.ObjectDef(
        objects.TestNest3,
        obj1,
        dryml.ObjectDef(
            objects.TestNest2,
            A=2),
        dryml.ObjectDef(
            objects.TestNest,
            dryml.ObjectDef(
                objects.TestNest2,
                A=5,)
            )
        )

    obj = obj_def.build()

    sel = dryml.Selector.build(obj_def)

    assert sel(obj, verbosity=2)


def test_selector_build_6():
    """
    Test that we can construct Selectors from nested objects
    """

    obj1 = objects.TestNest2(A=1)

    obj_def = dryml.ObjectDef(
        objects.TestNest3,
        obj1,
        dryml.ObjectDef(
            objects.TestNest2,
            A=2),
        dryml.ObjectDef(
            objects.TestNest,
            dryml.ObjectDef(
                objects.TestNest2,
                A=5,)
            )
        )

    obj = obj_def.build()

    sel = dryml.Selector.build(obj_def)

    assert sel(obj, verbosity=2)



import pytest
import dryml

import objects


@pytest.mark.usefixtures("create_temp_dir")
def test_compound_object_save_restore_1(create_temp_dir):
    # First create a definition object
    obj_def = dryml.ObjectDef(
        objects.TestNest,
        dryml.ObjectDef(
            objects.TestNest2,
            A=3
        )
    )

    # Next, create a compound object
    obj = obj_def.build()

    # Create object repo
    repo = dryml.Repo(directory=create_temp_dir)

    # Add object to the repo
    repo.add_object(obj)

    # Check that there are now two objects stored.
    assert len(repo) == 2

    # Save the objects
    repo.save()

    # Delete repo
    del repo

    # Recreate repo object
    repo = dryml.Repo(directory=create_temp_dir)

    # Check there are still two objects
    assert len(repo) == 2

    # Use original definition as a selector to grab the constructed object
    obj_ret = repo.get(selector=obj_def)

    assert type(obj_ret) is not list

    # Check new fetched object is the same as originally created object
    assert obj_ret.dry_id == obj.dry_id
    assert obj_ret.A.dry_id == obj.A.dry_id
    assert obj_ret.A.A == obj.A.A



from dryml.config import Meta
from dryml.object import Object
from dryml.models import Component
from dryml.models import Trainable as BaseTrainable
from dryml.models import TrainFunction as BaseTrainFunction
import zipfile
import torch


class Wrapper(Object):
    __dry_compute_context__ = "torch"

    @Meta.collect_args
    @Meta.collect_kwargs
    def __init__(self, cls, *args, **kwargs):
        if not isinstance(cls, type):
            raise TypeError(
                f"Expected first argument to be type. Got {type(cls)}")
        self.cls = cls
        self.args = args
        self.kwargs = kwargs
        self.obj = None

    def compute_prepare_imp(self):
        self.obj = self.cls(*self.args, **self.kwargs)

    def load_compute_imp(self, file: zipfile.ZipFile) -> bool:
        try:
            with file.open('state.pth', 'r') as f:
                self.obj.load_state_dict(torch.load(f))
            return True
        except Exception:
            return False

    def save_compute_imp(self, file: zipfile.ZipFile) -> bool:
        try:
            with file.open('state.pth', 'w') as f:
                torch.save(self.obj.state_dict(), f)
            return True
        except Exception:
            return False

    def compute_cleanup_imp(self):
        del self.obj
        self.obj = None


class TrainFunction(BaseTrainFunction):
    __dry_compute_context__ = 'torch'


class Model(Component):
    __dry_compute_context__ = 'torch'

    def prep_eval(self):
        pass

    def prep_train(self):
        pass


class Trainable(BaseTrainable):
    __dry_compute_context__ = 'torch'



from dryml.config import Meta
from dryml.models.component import Component
import xgboost
import tempfile
import zipfile


class Model(Component):
    @Meta.collect_kwargs
    def __init__(self, **kwargs):
        # It is subclass's responsibility to fill this
        # attribute with an actual keras class
        self.mdl = None
        self.mdl_kwargs = kwargs

    def compute_prepare_imp(self):
        self.mdl = self.cls(**self.mdl_kwargs)

    def compute_cleanup_imp(self):
        self.mdl = None

    def load_compute_imp(self, file: zipfile.ZipFile) -> bool:
        save_file_name = 'model.text'
        # Load Model
        if save_file_name not in file.namelist():
            # No model pickle file right now
            return True
        else:
            with file.open(save_file_name, 'r') as f:
                with tempfile.NamedTemporaryFile(
                        mode="w+b", suffix='.text') as temp_f:
                    # Write content to temp file
                    temp_f.write(f.read())
                    # Read temp file content
                    self.mdl.load_model(temp_f.name)
        return True

    def save_compute_imp(self, file: zipfile.ZipFile) -> bool:
        # Save Model
        if self.mdl is not None:
            with tempfile.NamedTemporaryFile(
                    mode="w+b", suffix='.text') as temp_f:
                self.mdl.save_model(temp_f.name)
                save_file_name = "model.text"
                with file.open(save_file_name, 'w') as f:
                    temp_f.seek(0)
                    f.write(temp_f.read())

        return True

    def __call__(self, X, *args, target=True, index=False, **kwargs):
        raise NotImplementedError()


class ClassifierModel(Model):
    def __init__(self):
        self.cls = xgboost.XGBClassifier

    def __call__(self, X, *args, target=True, index=False, **kwargs):
        return self.mdl.predict_proba(X, *args, **kwargs)


class RegressionModel(Model):
    def __init__(self):
        self.cls = xgboost.XGBRegressor

    def __call__(self, X, *args, target=True, index=False, **kwargs):
        return self.mdl.predict(X, *args, **kwargs)



from dryml.data.transforms import StaticTransform
from dryml.data.util import nestize


class ImageNormalization(StaticTransform):
    def numpy_eval(self, data, *args, **kwargs):
        import numpy as np
        return data.apply_X(
            nestize(lambda image: np.cast(image, np.float32) / 255.))

    def tf_eval(self, data, *args, **kwargs):
        import tensorflow as tf
        return data.apply_X(
            nestize(lambda image: tf.cast(image, tf.float32) / 255.))

    def torch_eval(self, data, *args, **kwargs):
        import torch
        return data.apply_X(
            nestize(lambda image: image.to(torch.float32) / 255.))



import pytest
import dryml
import objects_tf


def test_tfbase_1():
    """Test save/reload tfbase type objects"""

    test_obj = objects_tf.TestTF1()

    # Attempt to change class of the object.
    test_obj2 = dryml.change_object_cls(test_obj, objects_tf.TestTF1)

    assert test_obj2.definition() == test_obj.definition()


@pytest.mark.usefixtures("create_temp_named_file")
def test_tfbase_2(create_temp_named_file):
    """Test saving over an existing file"""

    test_obj = objects_tf.TestTF1()

    # Attempt to save model to file twice
    test_obj.save_self(create_temp_named_file)
    test_obj.save_self(create_temp_named_file)


def test_context_object_compute_check_tf_1():
    """
    Checks whether Eval behaves correctly when transformed to numpy
    """
    ray = pytest.importorskip('ray')
    import numpy as np
    import dryml.models.tf
    import tensorflow as tf

    num_classes = 5
    num_dims = 10

    model_def = dryml.ObjectDef(
        dryml.models.tf.keras.Trainable,
        model=dryml.ObjectDef(
            dryml.models.tf.keras.SequentialFunctionalModel,
            input_shape=(num_dims,),
            layer_defs=[
                ('Dense', {'units': 32, 'activation': 'relu'}),
                ('Dense', {'units': 32, 'activation': 'relu'}),
                ('Dense', {'units': num_classes, 'activation': 'softmax'}),
            ]),
        optimizer=dryml.ObjectDef(
            dryml.models.tf.Wrapper,
            tf.keras.optimizers.Adam),
        loss=dryml.ObjectDef(
            dryml.models.tf.Wrapper,
            tf.keras.losses.SparseCategoricalCrossentropy),
        train_fn=dryml.ObjectDef(
            dryml.models.tf.keras.BasicTraining,
            epochs=1))

    # Create data for eval
    batch_size = 20
    data = np.random.random((batch_size, num_dims))

    @ray.remote(num_cpus=1, num_gpus=0, max_calls=1)
    def test_method(model_def, data):
        model = model_def.build()

        import dryml.data
        ds = dryml.data.NumpyDataset(data)

        with dryml.context.ContextManager({'tf': {}}):
            result = model.eval(ds.batch(batch_size=batch_size)).numpy().peek()
            assert type(result) is np.ndarray

    ray.get(test_method.remote(model_def, data))


@pytest.mark.usefixtures("ray_server")
@pytest.mark.usefixtures("create_name")
def test_context_object_compute_check_tf_2(create_name):
    ray = pytest.importorskip('ray')
    import numpy as np
    import dryml.models.tf
    import tensorflow as tf

    num_classes = 5
    num_dims = 10

    model_def = dryml.ObjectDef(
        dryml.models.tf.keras.Trainable,
        model=dryml.ObjectDef(
            dryml.models.tf.keras.SequentialFunctionalModel,
            input_shape=(num_dims,),
            layer_defs=[
                ('Dense', {'units': 32, 'activation': 'relu'}),
                ('Dense', {'units': 32, 'activation': 'relu'}),
                ('Dense', {'units': num_classes, 'activation': 'softmax'}),
            ]),
        optimizer=dryml.ObjectDef(
            dryml.models.tf.Wrapper,
            tf.keras.optimizers.Adam),
        loss=dryml.ObjectDef(
            dryml.models.tf.Wrapper,
            tf.keras.losses.SparseCategoricalCrossentropy),
        train_fn=dryml.ObjectDef(
            dryml.models.tf.keras.BasicTraining,
            epochs=1))

    # Create data for eval
    batch_size = 20
    data = np.random.random((batch_size, num_dims))

    @ray.remote(num_cpus=1, num_gpus=0, max_calls=1)
    def test_method_1(model_def, data, dest_name):
        model = model_def.build()
        # Just created, there should be no model object
        assert model.model.mdl is None
        assert model.__dry_compute_data__ is None

        import dryml.data
        ds = dryml.data.NumpyDataset(data)

        print("test_method_1 info")
        print(f"ds batch size: {ds.batch_size}")
        print(f"batch_size: {batch_size}")
        print(f"peek: {ds.peek()}")

        with dryml.context.ContextManager({'tf': {}}):
            result = model.eval(ds.batch(batch_size=batch_size)).numpy().peek()

            assert model.model.mdl is not None
            assert model.__dry_compute_data__ is None

        assert model.model.mdl is None
        assert model.__dry_compute_data__ is not None

        model.save_self(dest_name)

        return result

    @ray.remote(num_cpus=1, num_gpus=0, max_calls=1)
    def test_method_2(data, dest_name):
        import dryml
        import dryml.models.tf
        model = dryml.load_object(dest_name)
        assert type(model) is dryml.models.tf.keras.Trainable

        # Just created, there should be no model object
        assert model.model.mdl is None
        # But we saved model data, so that should be there.
        assert model.__dry_compute_data__ is not None

        orig_compute_data = model.__dry_compute_data__

        import dryml.data
        ds = dryml.data.NumpyDataset(data)

        with dryml.context.ContextManager({'tf': {}}):
            result = model.eval(ds.batch(batch_size=batch_size)).numpy().peek()

            assert model.model.mdl is not None
            # We loaded the model into memory the
            # compute data should still be there though.
            assert model.__dry_compute_data__ is not None
            # It should also still be the same object
            assert id(orig_compute_data) == id(model.__dry_compute_data__)

        # We should have no model loaded anymore
        assert model.model.mdl is None
        assert model.__dry_compute_data__ is not None
        # Data should now be a different object. since we have to
        # write entirely new zip files
        assert id(orig_compute_data) != id(model.__dry_compute_data__)

        return result

    # Evaluate model and return result
    result_1 = ray.get(test_method_1.remote(model_def, data, create_name))
    result_2 = ray.get(test_method_2.remote(data, create_name))

    assert np.all(result_1 == result_2)



from collections import UserList, UserDict
from dryml.object import Object, ObjectDef
from dryml.config import Meta
from typing import Mapping


class List(Object, UserList):
    @Meta.collect_args
    def __init__(self, *args, **kwargs):
        objs = []
        for arg in args:
            if not isinstance(arg, Object):
                raise ValueError(
                    "List does not support elements of type"
                    f" {type(arg)}.")
            else:
                objs.append(arg)

        self.data.extend(objs)

    # We have to do a special implementation of definition
    # We want the reported dry_args to always match whats in
    # the list. this should be computed dynamically
    def definition(self):
        dry_args = []
        for obj in self:
            dry_args.append(obj.definition())
        return ObjectDef(
            type(self),
            *dry_args,
            dry_mut=True,
            **self.dry_kwargs)


class Tuple(Object):
    @Meta.collect_args
    def __init__(self, *args, **kwargs):
        objs = []
        for obj in args:
            if not isinstance(obj, Object):
                raise ValueError(f"Unsupported element of type: {type(obj)}")
            else:
                objs.append(obj)
        self.data = tuple(objs)

    def __getitem__(self, key):
        # Accessor
        return self.data[key]

    def __len__(self):
        return len(self.data)

    # We have to do a special implementation of definition
    # We want the reported dry_args to always match whats in
    # the list. this should be computed dynamically
    def definition(self):
        dry_args = []
        is_mutable = False
        for obj in self.data:
            obj_def = obj.definition()
            if obj_def.dry_mut:
                is_mutable = True
            dry_args.append(obj_def)
        return ObjectDef(
            type(self),
            *dry_args,
            dry_mut=is_mutable,
            **self.dry_kwargs)


class Dict(Object, UserDict):
    def __init__(
            self, in_dict: Mapping, **kwargs):
        for key in in_dict:
            self.data[key] = in_dict[key]

    # We have to do a special implementation of definition
    # We want the reported dry_args to always match whats in
    # the list. this should be computed dynamically
    def definition(self):
        # Build dry arg dictionary
        dry_arg = {}
        for key in self:
            obj = self[key]
            dry_arg[key] = obj.definition()
        return ObjectDef(
            type(self),
            dry_arg,
            dry_mut=True,
            **self.dry_kwargs)
